@@ -107,12 +107,12 @@ class BallTree:
         self.leaf_size = leaf_size
         self.built = 0
         self.size = 0
-        self.tree    = None
-        self.usage   = None
+        self.tree = None
+        self.usage = None
         self.sq_sums = None
-        self.order   = None
-        self.radii   = None
-        self.medians  = None
+        self.order = None
+        self.radii = None
+        self.medians = None
         # Assign the various pruning functions.
         self._balltree = ball_tree
         self._inner = prune.inner
@@ -138,7 +138,7 @@ class BallTree:
         if (self.tree is not None):
             # If possible, pack the points into available space.
             if (points.shape[1] <= self.tree.shape[1] - self.size):
-                self.tree[:,self.size:self.size+point.shape[1]] = points
+                self.tree[:,self.size:self.size+points.shape[1]] = points
                 self.size += points.shape[1]
             else:
                 # Store the old internal tree-defining attributes.
@@ -148,38 +148,37 @@ class BallTree:
                 old_order = self.order[:self.size]
                 old_radii = self.radii[:self.built]
                 old_medians = self.medians[:self.built]
-                old_size = self.size
                 # Compute the new size and initialize the memory for the new tree.
+                to_allocate = max(2 * self.size, self.size + points.shape[1])
                 self.size += points.shape[1]
-                self.tree = np.zeros(
-                    (self.tree.shape[0], self.size),
-                    dtype='float32', order='F')
+                self.tree = np.zeros((self.tree.shape[0], to_allocate),
+                                     dtype='float32', order='F')
                 # Assign the relevant internals for this tree.
-                self.usage   = np.zeros(self.tree.shape[1],  dtype='int64')
-                self.sq_sums = np.zeros(self.tree.shape[1],  dtype='float32')
-                self.order   = np.arange(self.tree.shape[1], dtype='int64') + 1
-                self.radii   = np.zeros(self.tree.shape[1],  dtype='float32')
-                self.medians  = np.zeros(self.tree.shape[1],  dtype='float32')
+                self.usage = np.zeros(self.tree.shape[1], dtype='int64')
+                self.sq_sums = np.zeros(self.tree.shape[1], dtype='float32')
+                self.order = np.arange(self.tree.shape[1], dtype='int64') + 1
+                self.radii = np.zeros(self.tree.shape[1], dtype='float32')
+                self.medians = np.zeros(self.tree.shape[1], dtype='float32')
                 # Pack the old points and the new points into a single tree
                 #  while ensuring the built parts of the previous tree are still valid.
                 self.tree[:,:old_tree.shape[1]] = old_tree
-                self.tree[:,old_tree.shape[1]:] = points
-                self.usage[:old_size] = old_usage
-                self.sq_sums[:old_size] = old_sq_sums
-                self.order[:self.built] = old_order
-                self.radii[:self.built] = old_radii
-                self.medians[:self.built] = old_medians
+                self.tree[:,old_tree.shape[1]:self.size] = points
+                self.usage[:old_usage.size] = old_usage
+                self.sq_sums[:old_sq_sums.size] = old_sq_sums
+                self.order[:old_order.size] = old_order
+                self.radii[:old_radii.size] = old_radii
+                self.medians[:old_medians.size] = old_medians
                 # Delete the old tree components.
                 del old_tree, old_usage, old_sq_sums, old_order, old_radii, old_medians
         else:
             # Save the points internally as the tree.
-            self.tree    = np.asarray(points, order='F', dtype='float32')
+            self.tree = np.asarray(points, order='F', dtype='float32')
             # Assign the relevant internals for this tree.
-            self.usage   = np.zeros(self.tree.shape[1],  dtype='int64')
-            self.sq_sums = np.zeros(self.tree.shape[1],  dtype='float32')
-            self.order   = np.arange(1, self.tree.shape[1]+1, dtype='int64')
-            self.radii   = np.zeros(self.tree.shape[1],  dtype='float32')
-            self.medians  = np.zeros(self.tree.shape[1],  dtype='float32')
+            self.usage = np.zeros(self.tree.shape[1], dtype='int64')
+            self.sq_sums = np.zeros(self.tree.shape[1], dtype='float32')
+            self.order = np.arange(1, self.tree.shape[1]+1, dtype='int64')
+            self.radii = np.zeros(self.tree.shape[1], dtype='float32')
+            self.medians = np.zeros(self.tree.shape[1], dtype='float32')
             # Store BallTree internals for knowing how to evaluate.
             self.built = 0
             self.size = self.tree.shape[1]
@@ -200,65 +199,88 @@ class BallTree:
         if (leaf_size is not None): self.leaf_size = leaf_size
         # Translate the root from python index to fortran index if provided.
         if (root is not None): root += 1
-        # Build tree (in-place operation).
-        #    .tree     will not be modified
-        #    .sq_sums  will contain the squared sums of each point
-        #    .radii    will be modified to have the radius of specific
-        #              node, or 0.0 if this point is a leaf.
-        #    .medians  will be modified to have the distance to the
-        #              median child node, or 0.0 if this point is a leaf.
-        #    .order    will be the list of indices (1-indexed) that
-        #              determine the structure of the ball tree.
-        self._balltree.build_tree(
-            self.tree, self.sq_sums,self.radii, self.medians, self.order[:self.size],
-            leaf_size=self.leaf_size, root=root,
-        )
+        # After this function completes, all points will be built.
         self.built = self.size
-        # Restructure the ball tree so the points are in locally contiguous blocks of
-        # memory (local by branch + leaf), as long as allowed (by user or memory).
-        if (reorder or ((reorder is None) and (self.tree.nbytes < self._balltree.max_copy_bytes))):
-            self._balltree.fix_order(self.tree, self.sq_sums, self.radii, self.medians, self.order[:self.built])
+        # Only call the routines if the tree has positive size.
+        if (self.size > 0):
+            # Build tree (in-place operation).
+            #    .tree     will not be modified
+            #    .sq_sums  will contain the squared sums of each point
+            #    .radii    will be modified to have the radius of specific
+            #              node, or 0.0 if this point is a leaf.
+            #    .medians  will be modified to have the distance to the
+            #              median child node, or 0.0 if this point is a leaf.
+            #    .order    will be the list of indices (1-indexed) that
+            #              determine the structure of the ball tree.
+            sq_dists = np.zeros(self.size, dtype='float32')
+            self._balltree.build_tree(
+                self.tree, self.sq_sums,self.radii, self.medians, sq_dists,
+                self.order[:self.size], leaf_size=self.leaf_size, root=root,
+            )
+            # Restructure the ball tree so the points are in locally contiguous blocks of
+            # memory (local by branch + leaf), as long as allowed (by user or memory).
+            if (reorder or ((reorder is None) and (self.tree.nbytes < self._balltree.max_copy_bytes))):
+                self._balltree.fix_order(self.tree, self.sq_sums, self.radii, self.medians, self.order[:self.built])
 
     # Find the "k" nearest neighbors to all points in z.
     def nearest(self, z, k=1, return_distance=True, transpose=True,
                 budget=None, randomness=None, include_unbuilt=True):
         # If only a single point was given, convert it to a matrix.
-        if (len(z.shape) == 1): z = z.reshape((1,-1))
+        if (len(z.shape) == 1):
+            singleton = True
+            z = z.reshape((1,-1))
+        else:
+            singleton = False
         # Transpose the points if appropriate.
         if transpose: z = z.T
         # Make sure the 'k' value isn't bigger than the tree size.
-        k = min(k, self.built)
+        k = min(k, self.size)
+        n = z.shape[1]
         # Initialize holders for output.
-        points  = np.asarray(z, order='F', dtype='float32')
+        points = np.asarray(z, order='F', dtype='float32')
         indices = np.ones((k, points.shape[1]), order='F', dtype='int64')
-        dists   = np.ones((k, points.shape[1]), order='F', dtype='float32')
+        dists = np.ones((k, points.shape[1]), order='F', dtype='float32')
+        iwork = np.ones((k+self.leaf_size+2, min(n,self._balltree.number_of_threads)), order='F', dtype='int64')
+        rwork = np.ones((k+self.leaf_size+2, min(n,self._balltree.number_of_threads)), order='F', dtype='float32')
         # Compute the nearest neighbors.
         self._balltree.nearest(
             points, k,
             self.tree, self.sq_sums, self.radii, self.medians, self.order[:self.built],
-            self.leaf_size, indices, dists, to_search=budget, randomness=randomness
+            self.leaf_size, indices, dists, iwork, rwork, to_search=budget, randomness=randomness
         )
+        del iwork, rwork
         if (budget is not None):
             budget = max(0, budget - self.built)
         # Compute distance to points that have not been built into the tree.
         if (self.size > self.built):
             uindices = np.zeros((k, points.shape[1]), order='F', dtype='int64')
-            udists   = np.zeros((k, points.shape[1]), order='F', dtype='float32')
-            uorder   = np.arange(self.size-self.built, dtype='int64') + 1
-            uleaf    = self.size - self.built # The "leaf size" for the unbuilt points.
+            udists = np.zeros((k, points.shape[1]), order='F', dtype='float32')
+            uorder = np.arange(self.size-self.built, dtype='int64') + 1
+            uleaf = self.size - self.built # The "leaf size" for the unbuilt points.
+            iwork = np.ones((k+uleaf+2, min(n,self._balltree.number_of_threads)), order='F', dtype='int64')
+            rwork = np.ones((k+uleaf+2, min(n,self._balltree.number_of_threads)), order='F', dtype='float32')
             self._balltree.nearest(
                 points, k,
                 self.tree[:,self.built:], self.sq_sums[self.built:], self.radii, self.medians,
-                uorder, uleaf, uindices, udists, to_search=budget,
+                uorder, uleaf, uindices, udists, iwork, rwork, to_search=budget,
             )
+            del iwork, rwork
             # Keep the nearest of the provided dists and indices.
-            print("uindices: ", uindices)
-            print("udists:   ", udists)
-            print()
-            raise(NotImplementedError("Need to merge results from ubuilt and built."))
+            indices = np.concatenate((indices, uindices+self.built), axis=0)
+            dists = np.concatenate((dists, udists), axis=0)
+            to_keep = np.argsort(dists, axis=0)
+            for i in range(points.shape[1]):
+                indices[:k,i] = indices[to_keep[:k,i],i]
+                dists[:k,i] = dists[to_keep[:k,i],i]
+            # Only keep the top k.
+            indices = indices[:k,:]
+            dists = dists[:k,:]
         # Update the usage statistics for all points referenced in return values.
-        i = (indices-1).flatten()
-        self.usage += np.bincount(i, minlength=self.usage.size)
+        self._balltree.bincount(indices.flatten(), self.usage)
+        # Revert to singleton if that's what was provided.
+        if singleton:
+            dists = dists[:,0]
+            indices = indices[:,0]
         # Return the results.
         if return_distance:
             if transpose: return dists.T, indices.T - 1
@@ -335,7 +357,7 @@ class BallTree:
 
 if __name__ == "__main__":
     print("_"*70)
-    print(" TESTING AXY MODULE")
+    print(" TESTING BALL TREE MODULE")
 
     from tlux.plot import Plot
     from tlux.random import well_spaced_ball, well_spaced_box
@@ -347,9 +369,9 @@ if __name__ == "__main__":
         return (3*x + np.cos(8*x)/2 + np.sin(5*y))
 
     # Set the number of points, dimension, and random seed for repeatability.
-    n = 10
+    n = 200
     d = 2
-    seed = 2
+    seed = 0
     np.random.seed(seed)
 
     # Create the test plot.
@@ -360,11 +382,11 @@ if __name__ == "__main__":
     tree = BallTree(x, reorder=False)
 
     # Define a function that reproduces the function via application of nearest neighbor.
-    fy = lambda x: y[tree(x, return_distance=False, k=1)]
+    fy = lambda x: y[tree(x, return_distance=False, k=1, budget=200)]
     x_min_max = np.asarray([np.min(x,axis=0), np.max(x,axis=0)]).T
 
     # Generate a plot to visualize the result.
     p = Plot()
     p.add("data", *(x.T), y)
-    p.add_function("balltree", fy, *x_min_max, plot_points=10000, vectorized=True)
+    p.add_function("balltree", fy, *x_min_max, plot_points=5000, vectorized=True)
     p.show()
