@@ -51,6 +51,13 @@
 # TODO: Adding multiple frames where the first has no edges and the
 #       rest have edges causes all frames to look like first.
 # 
+# 
+# # Uncomment the following to execute this file in place without system path issues.
+# import os, sys
+# _this_dir = os.path.dirname(os.path.realpath(__file__))
+# while (_this_dir in sys.path):
+#     sys.path.remove(_this_dir)
+# 
 # --------------------------------------------------------------------
 
 import random, numbers, os, webbrowser, sys, re, tempfile
@@ -87,11 +94,13 @@ DEFAULT_GRADIENT = np.array([[  94.,   79.,  162.],
 
 # PALETTE SOURCE: colorlover as cl
 # PALETTE SOURCE: np.array(cl.to_numeric(cl.scales['5']['qual']['Set2']))
-PALETTE = np.array([[ 102.,  194.,  165.],
-                    [ 252.,  141.,   98.],
-                    [ 141.,  160.,  203.],
-                    [ 231.,  138.,  195.],
-                    [ 166.,  216.,   84.]])
+PALETTE = np.array([
+    [ 102.,  194.,  165.],
+    [ 252.,  141.,   98.],
+    [ 141.,  160.,  203.],
+    [ 166.,  216.,   84.],
+    [ 231.,  138.,  195.],
+])
 PALETTE = PALETTE**2
 PALETTE = PALETTE / np.max(PALETTE) * 255
 # Re-order the palette so that the colors appear better
@@ -555,7 +564,7 @@ class Plot:
     #                 matrix of points as row-vectors for faster execution.
     # 
     #  ... <standard "add" arguments with adjusted defaults> ...
-    def add_function(self, name, func=None, min_max_x=None, min_max_y=[],
+    def add_function(self, name, func=None, min_max_x=None, min_max_y=tuple(),
                      x_vals=None, y_vals=None,
                      grid_lines=True, plot_points=PLOT_POINTS,
                      vectorized=False, mode=None, plot_type=None,
@@ -1165,10 +1174,10 @@ class Plot:
     # 
     #  ... <any additional plotly.offline.plot keyword arguments> ...
     def plot(self, title=None, x_range=None, y_range=None,
-             z_range=None, fixed=True, show_legend=True, layout={},
-             aspect_mode='cube', legend={}, scene_settings={},
-             axis_settings={}, x_axis_settings={}, y_axis_settings={},
-             z_axis_settings={}, hovermode="closest", 
+             z_range=None, fixed=True, show_legend=True, layout=None,
+             aspect_mode='cube', legend=None, scene_settings=None,
+             axis_settings=None, x_axis_settings=None, y_axis_settings=None,
+             z_axis_settings=None, hovermode="closest", 
              camera_position=DEFAULT_CAMERA_POSITION, html=True,
              file_name=None, show=True, append=False, height=None,
              width=None, loop_duration=5, bounce=False,
@@ -1178,6 +1187,14 @@ class Plot:
              show_slider_labels=True, show_play_pause=True,
              autoplay=False, loop=False, loop_pause=0,
              **kwargs):
+        # Declare default values for dictionary types.
+        if (layout is None): layout = {}
+        if (legend is None): legend = {}
+        if (scene_settings is None): scene_settings = {}
+        if (axis_settings is None): axis_settings = {}
+        if (x_axis_settings is None): x_axis_settings = {}
+        if (y_axis_settings is None): y_axis_settings = {}
+        if (z_axis_settings is None): z_axis_settings = {}
         # Update title, and all plot axis ranges
         if title == None:
             title = self.title
@@ -1659,7 +1676,7 @@ def multiplot(plots, x_domains=None, y_domains=None, html=True,
             # Extract the annotations for this plot
             plot_annotations = plot['layout'].pop('annotations',[])
             # Handle 3D and 2D differently
-            if specs[r][c]['is_3d']:
+            if specs[r][c].get('is_3d',False):
                 counter_3d += 1
                 scene_name = 'scene' + str(counter_3d)
                 fig['layout'][scene_name].update(plot['layout']['scene'])
@@ -1691,7 +1708,7 @@ def multiplot(plots, x_domains=None, y_domains=None, html=True,
             if len(plot_annotations) > 0:
                 plot['layout']['annotations'] = plot_annotations
             # Remove the 'scene' if there is one left over
-            if specs[r][c]['is_3d']: fig['layout'].pop('scene','')
+            if specs[r][c].get('is_3d',False): fig['layout'].pop('scene','')
     # Set the height and width properties, compensate for plotly spacing aroung SVG
     if type(width) != type(None):  
         width += 139
@@ -1916,6 +1933,9 @@ def _animate(data, plot_layout, loop_duration, bounce, transition,
 # ================================================
 
 if __name__ == "__main__":
+    while (os.path.dirname(__file__) in sys.path):
+        sys.path.remove(os.path.dirname(__file__))
+
     print()
     print("Creating a demonstration of most of the available (and useful) features!")
     print()
