@@ -26,13 +26,18 @@ def setup():
 # Build the AXY model (compiled Fortran code).
 def build_axy():
     import os, fmodpy
-    # Get the directory for the AXY compiled source code.
-    _dependencies = ["random.f90", "matrix_operations.f90", "sort_and_select.f90", "axy.f90"]
+    # Define the output directory.
     _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "axy")
-    # _dependencies = ["random.f90", "matrix_operations.f90", "sort_and_select.f90", "axy.f90"]
-    # _dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "approximate", "axy")
+    # Import the RANDOM module, it is self contained and simple.
+    _random = fmodpy.fimport(
+        input_fortran_file = os.path.join(_dir, "random.f90"),
+        output_dir = _dir
+    )
+    # Import the AXY module, it has many dependencies and uses BLAS and LAPACK libraries.
+    #   Get the directory for the AXY compiled source code.
+    _dependencies = ["axy_random.f90", "axy_matrix_operations.f90", "axy_sort_and_select.f90", "axy.f90"]
     _path = os.path.join(_dir, "axy.f90")
-    _kwargs = dict(
+    _axy_kwargs = dict(
         input_fortran_file = _path,
         dependencies = _dependencies,
         output_dir = _dir,
@@ -42,13 +47,13 @@ def build_axy():
         wrap = True,
         verbose = False,
     )
-    # Try and build using (local) defaults for fmodpy compilation.
+    #   Try and build using (local) defaults for fmodpy compilation.
     try:
-        _axy = fmodpy.fimport(**_kwargs)
+        _axy = fmodpy.fimport(**_axy_kwargs)
         _ = _axy.AXY()
-    # When local default fail, try specifying exactly what shared library functions are needed.
+    #   When local default fail, try specifying exactly what shared library functions are needed.
     except:
-        _axy = fmodpy.fimport( **_kwargs,
+        _axy = fmodpy.fimport( **_axy_kwargs,
             link_blas="", link_lapack="", rebuild=True,
             libraries = [_dir] + fmodpy.config.libraries,
             symbols = [
@@ -57,7 +62,8 @@ def build_axy():
                 ("omp_get_max_threads", "omp")
             ],
         )
-    return _axy
+    # Return the AXY and RANDOM modules.
+    return _axy, _random
 
 
 # Build the BallTree code.
